package midp.shopper;

import javax.microedition.midlet.*;
import javax.microedition.io.*;
import java.util.Vector;
import java.io.*;

public class HttpCommunicator implements Communicator { 
    String  serverAddress;
    boolean importing;
    boolean running;
    boolean trace;

    static final String DEFAULT_SERVER_ADDRESS = "www.garret.ru:8008";

    static final int MAX_CONTENT_LENGTH = 256;

    HttpCommunicator() {                                                    
        serverAddress = Shopper.getInstance().getAppProperty("SERVER_ADDRESS");
        if (serverAddress == null) { 
            serverAddress = DEFAULT_SERVER_ADDRESS;
        }
        trace = true;
    }

    public String name() { 
        return "HTTP";
    }

    public synchronized void start() {
        if (!running) { 
            running = true;
            new Thread(this).start();
        }
    }

    public synchronized void stop() {
        running = false;
    }

    public void send(final String phone, final String message) { 
        new Thread(new Runnable() {
             public void run() { 
                 sendMessage(phone, message); 
             }
        }).start();
    }

    private synchronized void sendMessage(String phone, String message) {
        try { 
            HttpConnection con = (HttpConnection) Connector.open("http://" + serverAddress);
            con.setRequestMethod(HttpConnection.POST);
            con.setRequestProperty("From", Shopper.getInstance().settings.selfPhone);
            con.setRequestProperty("To", phone);
            OutputStream out = con.openOutputStream(); 
            out.write(message.getBytes());
            InputStream in = con.openInputStream(); 
            if (trace) { 
                System.out.println("Send POST request to the server:");
                System.out.println(message);
                System.out.println("--------------------");
            }
            int rc = con.getResponseCode();
            if (trace) { 
                System.out.println("POST response: " + rc);
            }
            in.close();                
            out.close();
            con.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public synchronized void importCompleted() { 
        importing = false;
        notify();
    }
            
    static String readLine(InputStream in) throws IOException { 
        StringBuffer buf = new StringBuffer();
        int ch;
        while ((ch = in.read()) > 0) { 
            if (ch == '\n') { 
                System.out.println(buf.toString());
                return buf.toString();
            }
            buf.append((char)ch);
        }
        System.out.println("ch=" + ch);
        return buf.length() == 0 ? null : buf.toString();
    }
    
    public synchronized void run() {
        while (running) {
            if (!importing) {
                try { 
                    HttpConnection con = (HttpConnection) Connector.open("http://" + serverAddress);
                    con.setRequestMethod(HttpConnection.GET);
                    con.setRequestProperty("From", Shopper.getInstance().settings.selfPhone);
                    InputStream in = con.openInputStream();
                    if (trace) { 
                        System.out.println("Send GET request to the server");
                    }
                    int rc = con.getResponseCode();
                    if (trace) { 
                        System.out.println("GET response: " + rc);
                    }
                    if (rc == HttpConnection.HTTP_OK) { 
                        String from = Shopper.getInstance().findSender(con.getHeaderField("From"));
                        String s = readLine(in);
                        if ("<p>".equals(s)) { 
                            Vector v = new Vector();                           
                            while ((s = readLine(in)) != null && !"</p>".equals(s)) {
                                v.addElement(s);
                            }
                            if (!("</p>".equals(s) && "".equals(readLine(in)))) {
                                System.out.println("Bad request format");
                            }
                            importing = true;
                            if (trace) { 
                                System.out.println("Get product list from " + from);
                            }
                            new SyncForm(from, v);
                        } else if ("<s>".equals(s)) { 
                            String selection = readLine(in);
                            if (!("</s>".equals(readLine(in)) && "".equals(readLine(in)))) {
                                System.out.println("Bad request format");
                            }
                            importing = true;
                            if (trace) { 
                                System.out.println("Get need list from " + from);
                            }
                            new ImportForm(from, selection);
                        } else { 
                            System.out.println("Unexpected message body: " + s);
                        }
                    }
                    in.close();
                    con.close();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
            try { 
                wait(Shopper.getInstance().settings.pollInterval*60*1000);
            } catch (InterruptedException x) { 
                return;
            }
        }
    }
}

