package midp.shopper;

import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;

import javax.microedition.rms.*;
import java.util.Vector;
import java.util.Hashtable;
import java.io.IOException;

public class Shopper extends MIDlet implements CommandListener   { 
    private static Shopper instance = null;

    private Vector      categories;
    private RecordStore store;
    private RecordStore addressBookStore;
    private RecordStore settingsStore;
    
    private static Image[]   categoryImages;
    private static Image     productsImage;
    private static Hashtable imageHash;
    
    List           actionList;

    Vector         addressBook;
    Communicator[] communicators;

    Settings       settings;
    byte[]         selected;
    byte[]         origSelected;

    byte[]         predefined;
    int            nPredefined;

    static Command BACK_CMD;
    static Command ADD_CMD;
    static Command SELECT_CMD;
    static Command REMOVE_CMD;
    static Command RENAME_CMD;
    static Command ENTER_CMD;
    static Command SAVE_CMD;
    static Command QUIT_CMD;
    static Command SEND_CMD;
    static Command LIST_CMD;
    static Command MERGE_CMD;
    static Command IGNORE_CMD;
    static Command REPLACE_CMD;
    static Command ACCEPT_CMD;
    static Command REJECT_CMD;

    static final int SETTINGS_ID = 1;
    static final int PREDEFINED_ID = 1;
    static final int SELECTED_ID  = 2;

    static final String STORE_NAME = "Shopper207";
    static final String ADDRESS_BOOK_STORE_NAME = "ShopperAddressBook";
    static final String SETTINGS_STORE_NAME = "ShopperSettings";

    private Class[] actions = {
        getClassObject("midp.shopper.NeedList"),
        getClassObject("midp.shopper.CategoryList"),     
        getClassObject("midp.shopper.ResetList"),
        getClassObject("midp.shopper.SettingsForm"),
        getClassObject("midp.shopper.SendMessage")
    };

    public Vector getCategories() { 
        return categories;
    }

    /**
     * This method dynamically loads a class to work around a limitation
     * in the Forte compiler that prevents us from compiling ClassName.class
     * syntax.
     */
    public static Class getClassObject(String className) {
        try {
            return(Class.forName(className));
        }
        catch(Throwable err) {
            System.out.println(err.getClass().getName() + ": " +
                               err.getMessage());
            err.printStackTrace();
        }
        return(null);
    }


    static Shopper getInstance()
    {
        return(instance);
    }


    public Shopper() { 
        instance = this;
        String locale = getAppProperty("Shopper.Locale");
        if (locale == null || locale.length() == 0) { 
            locale = System.getProperty("microedition.locale");
        }        
        new Locale(locale);
        BACK_CMD   = new Command(Locale.current.back,  Command.BACK, 3);
        ADD_CMD    = new Command(Locale.current.add,   Command.SCREEN, 1);
        SELECT_CMD = new Command(Locale.current.select, Command.ITEM, 1);
        REMOVE_CMD = new Command(Locale.current.remove,Command.ITEM, 1);
        RENAME_CMD = new Command(Locale.current.rename,Command.ITEM, 1);
        ENTER_CMD  = new Command(Locale.current.enter, Command.SCREEN, 1);
        SAVE_CMD  = new Command(Locale.current.save, Command.SCREEN, 1);
        SEND_CMD  = new Command(Locale.current.send, Command.SCREEN, 1);
        LIST_CMD  = new Command(Locale.current.list, Command.SCREEN, 2);
        QUIT_CMD  = new Command(Locale.current.quit, Command.BACK, 3);
        MERGE_CMD = new Command(Locale.current.merge, Command.SCREEN, 1);
        REPLACE_CMD = new Command(Locale.current.replace, Command.SCREEN, 2);
        IGNORE_CMD  = new Command(Locale.current.ignore, Command.BACK, 3);
        ACCEPT_CMD = new Command(Locale.current.accept, Command.OK, 1);
        REJECT_CMD  = new Command(Locale.current.reject, Command.BACK, 2);

	categoryImages = new Image[Locale.current.categoryIcons.length];
        imageHash = new Hashtable();
	try 
	{
            productsImage = Image.createImage("/midp/shopper/products.png");
	    for (int i = 0; i < Locale.current.categoryIcons.length; i++) 
	    { 
		categoryImages[i] = Image.createImage("/midp/shopper/" + Locale.current.categoryIcons[i] + ".png");
                imageHash.put(Locale.current.categories[i], categoryImages[i]);
	    }
	} 
	catch (java.io.IOException err) 
	{
	    // ignore the image loading failure the application can recover.
	}
        communicators = CommunicatorFactory.getCommunicators();

        addressBook = new Vector();
        actionList = new List(Locale.current.name, List.IMPLICIT);
        actionList.addCommand(SELECT_CMD);
        actionList.addCommand(QUIT_CMD);
        actionList.setCommandListener(this);

        openSettings();
        openStore();
        openAddressBook();

        int nActions = Locale.current.actionNames.length;
        Communicator communicator = getCommunicator();
        if (communicator == null) { 
            nActions -= 1;
        } else { 
            communicator.start();
        }
        for (int i = 0; i < nActions; i++) { 
            actionList.append(Locale.current.actionNames[i], null);
        }    
    }

    public Communicator getCommunicator() {
        return settings.transport >= 0 ? communicators[settings.transport] : null;
    }

    private void openStore() {
        try { 
            try { 
                store = RecordStore.openRecordStore(STORE_NAME, false);
                loadData();
            } catch (RecordStoreNotFoundException x) {             
                store = RecordStore.openRecordStore(STORE_NAME, true);
                initializeStore();
            }            
            origSelected = new byte[selected.length];
            System.arraycopy(selected, 0, origSelected, 0, selected.length);
        } catch(Exception x) { 
            criticalException(x);
        }
    }

    private void openSettings() { 
        try { 
            try { 
                settingsStore = RecordStore.openRecordStore(SETTINGS_STORE_NAME, false);
                try { 
                    byte[] record =  settingsStore.getRecord(SETTINGS_ID);                    
                    settings = new Settings(record);
                    return;
                } catch (InvalidRecordIDException x) {}
            } catch (RecordStoreNotFoundException x) {             
                settingsStore = RecordStore.openRecordStore(SETTINGS_STORE_NAME, true);
            }            
            settings = new Settings();
            byte[] record = settings.pack();
            settingsStore.addRecord(record, 0, record.length);
        } catch(Exception x) { 
            criticalException(x);
        }
    }

   
    private void openAddressBook() {
        try { 
            addressBookStore = RecordStore.openRecordStore(ADDRESS_BOOK_STORE_NAME, true);
            int last = addressBookStore.getNextRecordID();
            for (int recordId = 0; ++recordId < last;) { 
                byte[] record = null;
                try { 
                    record = addressBookStore.getRecord(recordId);                    
                } catch (InvalidRecordIDException x) {
                    // record was deleted 
                }
                if (record != null) { 
                    addressBook.addElement(new Address(recordId, record));
                }
            }
        } catch (RecordStoreException x) {             
            criticalException(x);
        }
    }

    void saveSettings() { 
        byte[] record = settings.pack();
        try {
            settingsStore.setRecord(SETTINGS_ID, record, 0, record.length);
        } catch (RecordStoreException x) {             
            criticalException(x);
        }
    }


    void addAddress(String name, String phone) { 
        Address address = new Address(name, phone);
        byte[] record = address.pack();
        try {
            address.recordId = addressBookStore.addRecord(record, 0, record.length);
            addressBook.addElement(address);
        } catch (RecordStoreException x) {             
            criticalException(x);
        }
    }

    void removeAddress(int index) { 
        Address address = (Address)addressBook.elementAt(index);
        addressBook.removeElementAt(index);
        try {             
            addressBookStore.deleteRecord(address.recordId);      
        }  catch (RecordStoreException x) {             
            criticalException(x);
        }
    }

    private void initializeStore() {
        try { 
            categories = new Vector();
            int n = 0;
            for (int i = 0; i < Locale.current.categories.length; i++) {   
                n += 1;
                Category category = new Category(-n, Locale.current.categories[i], categoryImages[i]);
                categories.addElement(category);

                for (int j = 0; j < Locale.current.products[i].length; j++) { 
                    n += 1;
                    Product product = new Product(-n, category, Locale.current.products[i][j]);                
                    category.products.addElement(product);
                }
            }
            nPredefined = n+1;
            predefined = new byte[(n + 8) >> 3];
            store.addRecord(predefined, 0, predefined.length);
            selected = new byte[predefined.length];
            store.addRecord(selected, 0, selected.length);
        } catch(Exception x) { 
            criticalException(x);
        }
    }
        

    private void loadData() throws Exception { 
        Hashtable hash = new Hashtable();
        try { 
            predefined = store.getRecord(PREDEFINED_ID);
            selected = store.getRecord(SELECTED_ID);
            byte[] deleted = predefined;
            while (true) { 
                int n = 0;
                categories = new Vector();
                for (int i = 0; i < Locale.current.categories.length; i++) {   
                    Category category = null;
                    n += 1;
                    if ((n>>3) >= deleted.length || (deleted[n>>3] & (1 << (n & 7))) == 0) { 
                        category = new Category(-n, Locale.current.categories[i], categoryImages[i]);
                        hash.put(new Integer(-n), category);
                        categories.addElement(category);
                    }
                    for (int j = 0; j < Locale.current.products[i].length; j++) { 
                        n += 1;
                        if (category != null && ((n>>3) >= deleted.length || (deleted[n>>3] & (1 << (n & 7))) == 0)) { 
                            Product product = new Product(-n, category, Locale.current.products[i][j]);                
                            category.products.addElement(product);
                        }
                    }
                }
                nPredefined = n+1;
                int delta = ((n + 8) >> 3) - deleted.length;
                if (delta > 0) { 
                    predefined = deleted = new byte[(n + 8) >> 3];
                    store.setRecord(PREDEFINED_ID, deleted, 0, deleted.length);
                    selected = new byte[selected.length + delta + 1];
                    store.setRecord(SELECTED_ID, selected, 0, selected.length);
                    continue;
                }
                break;
            }
            int last = store.getNextRecordID();
            for (int recordId = SELECTED_ID; ++recordId < last;) { 
                byte[] record = null;
                try { 
                    record = store.getRecord(recordId);                    
                } catch (InvalidRecordIDException x) {
                    // record was deleted 
                }
                if (record == null) { 
                    continue;
                }
                int categoryId = Convert.unpack(record);
                String name = Convert.bytes2Str(record, 4, (record.length - 4) >> 1);
                if (categoryId == 0) { 
                    Category category = new Category(recordId, name, productsImage);
                    categories.addElement(category);
                    hash.put(new Integer(recordId), category);
                } else { 
                    Category category = (Category)hash.get(new Integer(categoryId));
                    if (category == null) { 
                        criticalError("Failed to locate category");
                        continue;
                    }
                    Product product = new Product(recordId, category, name);
                    category.products.addElement(product);
                }
            }
        } catch (Exception x) {
            try { 
                store.closeRecordStore();
                RecordStore.deleteRecordStore(STORE_NAME);
            } catch (Exception xx) {}
            criticalException(x);
        }
    }

    public void broadcastProductList() {
        Communicator communicator = getCommunicator();
        if (communicator != null) { 
            StringBuffer buf = new StringBuffer();
            buf.append("<p>\n");
            try { 
                int last = store.getNextRecordID();
                for (int recordId = SELECTED_ID; ++recordId < last;) { 
                    byte[] record = null;
                    try { 
                        record = store.getRecord(recordId);                    
                    } catch (InvalidRecordIDException x) {
                        // record was deleted 
                    }
                    if (record == null) { 
                        buf.append('-');
                    } else { 
                        buf.append(Convert.bytes2Hex(record));
                    }
                    buf.append('\n');
                }
                buf.append("</p>\n\n");
                String message = buf.toString();
                
                for (int i = addressBook.size(); --i >= 0;) { 
                    Address address = (Address)addressBook.elementAt(i);
                    communicator.send(address.phone, message);
                }
            } catch (RecordStoreException x) {             
                criticalException(x);
            }
        }
    }

    public void syncProductList(Vector v) {
        try { 
            int last = store.getNextRecordID();
            int size = v.size();
            int recordId = 0;
            for (int i = store.getNextRecordID()-SELECTED_ID-1; i < size; i++) { 
                String str = (String)v.elementAt(i);
                if (str.equals("-")) { 
                    recordId = store.addRecord(new byte[0], 0, 0);
                    store.deleteRecord(recordId);
                } else { 
                    byte[] record = Convert.hex2Bytes(str);
                    recordId =  store.addRecord(record, 0, record.length);
                    int categoryId = Convert.unpack(record);
                    String name = Convert.bytes2Str(record, 4, (record.length - 4) >> 1);
                    if (categoryId == 0) { 
                        Category category = new Category(recordId, name, productsImage);
                        categories.addElement(category);
                    } else { 
                        for (int j = categories.size(); --j >= 0;) { 
                            Category category = (Category)categories.elementAt(j);
                            if (category.recordId == categoryId) { 
                                Product product = new Product(recordId, category, name);
                                category.products.addElement(product);
                                break;
                            } 
                        }
                    }
                }
                if (recordId != i+SELECTED_ID+1) { 
                    System.out.println("Unexpected record ID: " + recordId + " vs. " + (i+SELECTED_ID+1));
                    break;
                }
            }
            if (recordId > 0) { 
                int i = (nPredefined + recordId + 7) >> 3;
                while (i >= selected.length) { 
                    byte[] newSelected = new byte[selected.length*2];
                    System.arraycopy(selected, 0, newSelected, 0, selected.length);
                    selected = newSelected;
                    store.setRecord(SELECTED_ID, newSelected, 0, newSelected.length);
                }
            }
        } catch (RecordStoreException x) {             
            criticalException(x);
        }
    }            

    public void storeCategory(Category category) {
        try { 
            byte[] name = Convert.str2Bytes(category.name);
            byte[] record = new byte[name.length + 4];
            System.arraycopy(name, 0, record, 4, name.length);
            store.setRecord(category.recordId, record, 0, record.length);
        } catch (Exception x) {
            criticalException(x);
        }
    }

    public Category addCategory(String name) { 
        byte[] nameBytes = Convert.str2Bytes(name);
        byte[] record = new byte[nameBytes.length + 4];
        int recordId = 0;
        System.arraycopy(nameBytes, 0, record, 4, nameBytes.length);
        try { 
            recordId = store.addRecord(record, 0, record.length);
        } catch (Exception x) { 
            criticalException(x);
        }
        Category category = new Category(recordId, name, productsImage);
        categories.addElement(category);
        return category;
    }

    public boolean removeCategory(int i) { 
        Category category = (Category)categories.elementAt(i);
        if (category.products.size() != 0) { 
            error(Locale.current.notEmpty);
            return false;
        } else { 
            if (category.isPredefined()) { 
                int j = -category.recordId;
                predefined[j>>3] |= (byte)(1 << (j & 7));
                try { 
                    store.setRecord(PREDEFINED_ID, predefined, 0, predefined.length);            
                } catch (Exception x) { 
                    criticalException(x);
                }
            } else { 
                try { 
                    store.deleteRecord(category.recordId);            
                } catch (Exception x) { 
                    criticalException(x);
                }
            }
            categories.removeElementAt(i);
            return true;
        }
    }

    public Product addProduct(String name, Category category) { 
        byte[] nameBytes = Convert.str2Bytes(name);
        byte[] record = new byte[nameBytes.length + 4];
        int recordId = 0;
        System.arraycopy(nameBytes, 0, record, 4, nameBytes.length);
        Convert.pack(record, category.recordId);        
        try { 
            recordId = store.addRecord(record, 0, record.length);
        } catch (Exception x) { 
            criticalException(x);
        }
        Product product = new Product(recordId, category, name);
        category.products.addElement(product);
        int i = (nPredefined + recordId + 7) >> 3;
        while (i >= selected.length) { 
            byte[] newSelected = new byte[selected.length*2];
            System.arraycopy(selected, 0, newSelected, 0, selected.length);
            selected = newSelected;
            try { 
                store.setRecord(SELECTED_ID, newSelected, 0, newSelected.length);
            } catch (Exception x) { 
                criticalException(x);
            }
        }
        return product;
    }

    public void removeProduct(int i, Product product) { 
        if (product.isPredefined()) { 
            int j = -product.recordId;
            predefined[j>>3] |= (byte)(1 << (j & 7));
            try { 
                store.setRecord(PREDEFINED_ID, predefined, 0, predefined.length);            
            } catch (Exception x) { 
                criticalException(x);
            }
        } else { 
            try { 
                store.deleteRecord(product.recordId);            
            } catch (Exception x) { 
                criticalException(x);
                return;
            }
        }
        product.category.products.removeElementAt(i);
    }
        
    public void restoreData() { 
        System.arraycopy(origSelected, 0, selected, 0, origSelected.length);
        for (int i = origSelected.length; i < selected.length; i++) { 
            selected[i] = 0;
        }
        saveSelected();
    }    

    public void clearData() { 
        byte[] sel = selected;
        for (int i = 0, n = sel.length; i < n; i++) { 
            sel[i] = 0;
        }
        saveSelected();
    }

    public void setData() { 
        byte[] sel = selected;
        for (int i = 0, n = sel.length; i < n; i++) { 
            sel[i] = (byte)-1;
        }
        saveSelected();
    }

    protected void destroyApp(boolean unconditional) {
        if (store != null) { 
            try { 
                store.closeRecordStore();
            } catch (Exception x) {
                criticalException(x);
            }
        }
        if (addressBookStore != null) { 
            try { 
                addressBookStore.closeRecordStore();
            } catch (Exception x) {
                criticalException(x);
            }
        }
        if (settingsStore != null) { 
            try { 
                settingsStore.closeRecordStore();
            } catch (Exception x) {
                criticalException(x);
            }
        }
        actionList = null;
    }    

    protected  void pauseApp() {
    }

    protected void startApp() {
        Display.getDisplay(this).setCurrent(actionList);
    }

    public static void goBack() {
        instance.startApp();
    }

    public void merge(byte[] vector) { 
        int n = selected.length < vector.length ? selected.length : vector.length;
        for (int i = 0; i < n; i++) { 
            selected[i] |= vector[i];
        }
        saveSelected();         
    }

    public String getNeedList() {
        return "<s>\n" + Convert.bin2Hex(selected) + "\n</s>\n\n";
    }
 

    public void commandAction(Command c, Displayable d) {
        if (c == QUIT_CMD) { 
            destroyApp(true);
            notifyDestroyed();
        } else { 
            int pos = actionList.getSelectedIndex();
            try {
                Displayable current = (Displayable)actions[pos].newInstance();
                Display.getDisplay(this).setCurrent(current);
            }
            catch(Exception err) {
                criticalException(err);
            }
        }
    }

    void criticalException(Exception x) {         
        x.printStackTrace();
        criticalError(x.toString());
    }
    
    void criticalError(String msg) {
        System.err.println(msg);
        error(msg);
    }

    boolean isSelected(Product product) { 
        int i = nPredefined + product.recordId;
        return (selected[i>>3] & (1 << (i & 7))) != 0;
    }

    boolean setSelected(Product product, boolean select) { 
        int i = nPredefined + product.recordId;
        if (((selected[i>>3] & (1 << (i & 7))) != 0) ^ select) { 
            selected[i>>3] ^= (1 << (i & 7));
            return true;
        }
        return false;
    }

    void saveSelected() { 
         try { 
            store.setRecord(SELECTED_ID, selected, 0, selected.length);            
        } catch (Exception x) { 
            criticalException(x);
            return;
        }
    }

    void error(String msg) {
        Alert a = new Alert(Locale.current.error);
        a.setType(AlertType.ERROR);
        a.setString(msg);
        a.setTimeout(Alert.FOREVER);
        Display.getDisplay(this).setCurrent(a);
    }

    String findSender(String phone) { 
        for (int i = addressBook.size(); --i >= 0;) { 
            Address addr = (Address)addressBook.elementAt(i);
            if (phone.endsWith(addr.phone)) { 
                return addr.name;
            }
        }
        return phone;
    }
}
