package midp.shopper;

import javax.microedition.midlet.*;
import javax.wireless.messaging.*;
import javax.microedition.io.*;
import java.io.IOException;
import java.util.Vector;

public class SmsCommunicator implements Communicator { 
    MessageConnection smsconn;
    String            smsPort;
    boolean           importing;
    boolean           running;

    static final String DEFAULT_SMS_PORT = "12345";

    public SmsCommunicator() { 
        smsPort = Shopper.getInstance().getAppProperty("SMS_PORT");
        if (smsPort == null) { 
            smsPort = DEFAULT_SMS_PORT;
        }
        String smsConnection = "sms://:" + smsPort;
        try {
            smsconn = (MessageConnection) Connector.open(smsConnection);
            // smsconn.setMessageListener(this);
        } catch (IOException ioe) {
            ioe.printStackTrace();
        }
        try {
            PushRegistry.registerConnection(smsConnection,
                                            getClass().getName(), 
                                            "*");
        } catch ( ClassNotFoundException cnf ) {            
        } catch ( IOException ioe ) {
        }
    }

    public void start() {
        if (!running) { 
            running = true;
            new Thread(this).start();
        }
    }

    public void stop() {
        running = false;
    }

    public String name() { 
        return "SMS";
    }

    public void send(final String phone, final String message) { 
        new Thread(new Runnable() { 
            public void run() { 
                MessageConnection smsconn = null;
                try {
                    String addr = "sms://" + phone + ':' + smsPort;
                    smsconn = (MessageConnection) Connector.open(addr);
                    TextMessage msg =
                        (TextMessage)smsconn.newMessage(MessageConnection.TEXT_MESSAGE);
                    msg.setPayloadText(message);
                    smsconn.send(msg);
                } catch (Throwable t) {
                    System.out.println("Send caught: ");
                    t.printStackTrace();
                }
                if (smsconn != null) {
                    try {
                        smsconn.close();
                    } catch (IOException ioe) {
                        System.out.println("Closing connection caught: ");
                        ioe.printStackTrace();
                    }                
                }
            }
        }).start();
    }

    public synchronized void importCompleted() { 
        importing = false;
        notify();
    }
            

    public void run() {
        while (true) { 
            try { 
                synchronized(this) { 
                    if (importing) { 
                        wait();
                    }
                }
            } catch (Exception x) {}

            Message msg = null;
            try {
                msg = smsconn.receive();
            } catch (IOException e) {
                if (running) { 
                    e.printStackTrace();
                }
            }
            if (msg instanceof TextMessage) {
                String text = ((TextMessage)msg).getPayloadText();
                String from = Shopper.getInstance().findSender(msg.getAddress());                
                if (text.startsWith("<s>\n")) { 
                    int end = text.indexOf("</s>");
                    if (end < 0) { 
                        System.out.println("Bad request format");
                    } else {                     
                        synchronized(this) { 
                            importing = true;
                        } 
                        String selection = text.substring(4, end);                    
                        new ImportForm(from, selection);
                    }
                } else if (text.startsWith("<p>\n")) { 
                    Vector v = new Vector();
                    int p, q = 4;
                    while ((p = text.indexOf('\n', q)) >= 0) { 
                        String str = text.substring(p, q);
                        if ("</p>".equals(str)) { 
                            break;
                        }
                        v.addElement(str);
                        q = p + 1;
                    }
                    synchronized(this) { 
                        importing = true;
                    } 
                    new SyncForm(from, v);
                } else { 
                    System.out.println("Unexpected message body: " + text);
                }
            }
        }
    }
}

